/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
/*
 * Copyright (c) 1988 by Sun Microsystems, Inc.
 */

/*
 * Command to change one's public key in the public key database
 */
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpcsvc/ypclnt.h>
#include <sys/file.h>
#include <stdio.h>
#include <pwd.h>
#include <mp.h>

#include "rpcsec.h"

static char *domain;

static void	usage(void);
static void	getseed(char *seed, int seedsize, unsigned char *pass);
static void	genkeys(char *public, char *secret, char *pass);
static void	adjust(char *keyout, char *keyin);
static int	setpublicmap(char *name, char *public, char *secret);
static struct passwd *ypgetpwuid(int uid);

int
main(int argc, char **argv)
{
	char name[MAXNETNAMELEN+1];
	char public[HEXKEYBYTES + 1];
	char secret[HEXKEYBYTES + 1];
	char crypt1[HEXKEYBYTES + KEYCHECKSUMSIZE + 1];
	char crypt2[HEXKEYBYTES + KEYCHECKSUMSIZE + 1];
	int status;	
	char *pass;
	struct passwd *pw;
	char *master;
	int euid;
	int force;
	int c;

	force = 0;

	while ((c = getopt(argc, argv, "f")) != EOF) {
		switch (c) {
		case 'f':
			force = 1;
			break;
		default:
			usage();
		}
	}

	yp_get_default_domain(&domain);
	if (yp_master(domain, PUBLICKEY_MAP, &master) != 0) {
		fprintf(stderr, "can't find master of publickey database\n");
		return 1;
	}

	if (optind == argc - 1) {
		if (!(pw = getpwnam(argv[optind]))) {
			fprintf(stderr, "No passwd entry for user %s\n",
				argv[optind]);
			return 1;
		}
		if ((euid = pw->pw_uid) != 0) {
			user2netname(name, euid, NULL);
		} else {
			host2netname(name, NULL, NULL);
		}
	} else if (optind == argc) {
		getnetname(name);
		euid = geteuid();
	} else {
		usage();
	}
	fprintf(stderr, "Generating new key for %s.\n", name);

	pw = euid? ypgetpwuid(euid) : getpwuid(0);
	if (!pw) {
		fprintf(stderr, "No %spassword found: can't change key.\n",
			euid? "" : "NIS ");
		return 1;
	}

	pass = getpass("Password:");
	if (!force) {
		if (strcmp(crypt(pass, pw->pw_passwd), pw->pw_passwd) != 0) {
			fprintf(stderr, "Invalid password.\n");
			return 1;
		}
	}

	genkeys(public, secret, pass);	

	bcopy(secret, crypt1, HEXKEYBYTES);
	bcopy(secret, crypt1 + HEXKEYBYTES, KEYCHECKSUMSIZE);
	crypt1[HEXKEYBYTES + KEYCHECKSUMSIZE] = 0;
	xencrypt(crypt1, pass);

	if (force) {
		bcopy(crypt1, crypt2, HEXKEYBYTES + KEYCHECKSUMSIZE + 1);	
		xdecrypt(crypt2, getpass("Retype password:"));
		if (bcmp(crypt2, crypt2 + HEXKEYBYTES, KEYCHECKSUMSIZE) != 0 ||
	    	    bcmp(crypt2, secret, HEXKEYBYTES) != 0) {	
			fprintf(stderr, "Password incorrect.\n");
			exit(1);
		}
	}

	fprintf(stderr, "Sending key change request to %s...\n", master);
	status = setpublicmap(name, public, crypt1);
	if (status != 0) {
		fprintf(stderr,
			"chkey: unable to update yp database (%u): %s\n"
			"Perhaps %s is down?\n",
			status, yperr_string(status), master);
		exit(1);
	}
	fprintf(stderr, "Done.\n");

	if (key_setsecret(secret) < 0) {
		fprintf(stderr, "Unable to login with new secret key.\n");
		exit(1);
	}

	return 0;
}

static void
usage(void)
{
	fprintf(stderr, "usage: chkey [-f]\n");
	exit(1);
}


/*
 * Generate a seed. This is not a very good seed.
 */
#ifndef MUST_USE_DEVRANDOM
static void
oldgetseed(char *seed, int seedsize, unsigned char *pass)
{
	struct timeval	tv;
	int		i, rseed;

	gettimeofday(&tv, NULL);
	rseed = tv.tv_sec + tv.tv_usec;
	for (i = 0; i < 8; i++) {
		rseed ^= (rseed << 8) | pass[i];
	}
	srandom(rseed);

	for (i = 0; i < seedsize; i++) {
		seed[i] = (random() & 0xff) ^ pass[i % 8];
	}
}
#endif

static void
getseed(char *seed, int seedsize, unsigned char *pass)
{
	int	fd, n, got;

	if ((fd = open("/dev/random", O_RDONLY)) < 0) {
		perror("can't open /dev/random");
#ifdef MUST_USE_DEVRANDOM
		fprintf(stderr, "giving up.\n");
		exit(2);
#else
		oldgetseed(seed, seedsize, pass);
		return;
#endif
	}

	for (got = 0; got < seedsize; got += n) {
		if ((n = read(fd, seed+got, seedsize-got)) < 0) {
			fprintf(stderr, "failed to read from "
					"/dev/random after %d bytes\n",
					got);
			exit(2);
		}
	}
	close(fd);
}


/*
 * Generate a random public/secret key pair
 */
static void
genkeys(char *public, char *secret, char *pass)
{
	int i;
 
#define BASEBITS	(8*sizeof(short) - 1)
#define BASE		(1 << BASEBITS)
 
	MINT *pk = itom(0);
 	MINT *sk = itom(0);
	MINT *tmp;
	MINT *base = itom(BASE);
	MINT *root = itom(PROOT);
	MINT *modulus = xtom(HEXMODULUS);
	short r;
	unsigned short seed[KEYSIZE/BASEBITS + 1];
	char *xkey;

	getseed((char *)seed, sizeof(seed), (unsigned char *)pass);	
	for (i = 0; i < KEYSIZE/BASEBITS + 1; i++) {
		r = seed[i] % BASE;
		tmp = itom(r);
		mult(sk, base, sk);
		madd(sk, tmp, sk);
		mfree(tmp);  
	}
	tmp = itom(0);
	mdiv(sk, modulus, tmp, sk);
	mfree(tmp);
	pow(root, sk, modulus, pk); 
	xkey = mtox(sk);   
	adjust(secret, xkey);
	xkey = mtox(pk);
	adjust(public, xkey);
	mfree(sk);
	mfree(base);
	mfree(pk);
	mfree(root);
	mfree(modulus);
} 

/*
 * Adjust the input key so that it is 0-filled on the left
 */
static void
adjust(char *keyout, char *keyin)
{
	int	len, off;

	len = strlen(keyin);
	off = HEXKEYBYTES - len;
	memmove(keyout + off, keyin, len);
	memset(keyout, 0, off);
}

/*
 * Set the entry in the public key map
 */
static int
setpublicmap(char *name, char *public, char *secret)
{
	char	pkent[1024];
	
	sprintf(pkent, "%s:%s", public, secret);
	return yp_update(domain, PUBLICKEY_MAP, YPOP_STORE, 
				name, strlen(name), pkent, strlen(pkent));
}

static struct passwd *
ypgetpwuid(int uid)
{
	static struct passwd pw;
	char	uidstr[10];
	char	*val;
	int	vallen;
	char	*p;

	sprintf(uidstr, "%d", uid);
	if (yp_match(domain, "passwd.byuid", uidstr, strlen(uidstr), 
		     &val, &vallen)) {
		return NULL;
	}
	if (!(p = strchr(val, ':')))
		return NULL;

	pw.pw_passwd = p + 1;
	if (!(p = strchr(pw.pw_passwd, ':')))
		return NULL;
	*p = 0;
	return (&pw);
}
